package cn.iocoder.yudao.module.reward.api.reward;

import cn.hutool.core.bean.BeanUtil;
import cn.hutool.core.lang.generator.SnowflakeGenerator;
import cn.iocoder.yudao.framework.redis.helper.RedisDistributedLock;
import cn.iocoder.yudao.module.member.api.score.MemberUserScoreApi;
import cn.iocoder.yudao.module.member.api.score.dto.MemberUserScoreOperateReqDTO;
import cn.iocoder.yudao.module.member.api.user.MemberUserApi;
import cn.iocoder.yudao.module.member.api.user.dto.UserRespDTO;
import cn.iocoder.yudao.module.member.enums.ScoreSourceTypeEnum;
import cn.iocoder.yudao.module.reward.api.reward.dto.RedeemRewardReqVO;
import cn.iocoder.yudao.module.reward.api.reward.dto.RedeemRewardRespDTO;
import cn.iocoder.yudao.module.reward.dal.dataobject.redeem.RewardRedeemDO;
import cn.iocoder.yudao.module.reward.dal.dataobject.reward.RewardDO;
import cn.iocoder.yudao.module.reward.dal.mysql.redeem.RewardRedeemMapper;
import cn.iocoder.yudao.module.reward.dal.mysql.reward.RewardMapper;
import cn.iocoder.yudao.module.reward.enums.RewardRedeemStatusEnum;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.validation.annotation.Validated;

import javax.annotation.Resource;
import java.util.*;

import static cn.iocoder.yudao.framework.common.exception.util.ServiceExceptionUtil.exception;
import static cn.iocoder.yudao.module.infra.enums.ErrorCodeConstants.GET_LOCK_FAILED;
import static cn.iocoder.yudao.module.member.enums.ErrorCodeConstants.USER_NOT_EXISTS;
import static cn.iocoder.yudao.module.reward.enums.ErrorCodeConstants.*;

@Service
@Validated
public class RedeemRewardApiImpl implements RedeemRewardApi {

    @Resource
    private RewardMapper rewardMapper;
    @Resource
    private RewardRedeemMapper rewardRedeemMapper;
    @Resource
    private MemberUserApi memberUserApi;
    @Resource
    private RedisDistributedLock redisDistributedLock;
    @Resource
    private MemberUserScoreApi memberUserScoreApi;
    @Resource
    private SnowflakeGenerator snowflakeGenerator;

    @Override
    @Transactional(rollbackFor = Exception.class)
    public RedeemRewardRespDTO redeemReward(RedeemRewardReqVO redeemRewardReqVO) {
        // 查询礼品
        RewardDO rewardDO = rewardMapper.selectById(redeemRewardReqVO.getRewardId());
        if (rewardDO == null) {
            throw exception(REWARD_NOT_EXISTS);
        }
        // 礼品未开启
        if (rewardDO.getStatus() != 1) {
            throw exception(REWARD_NOT_ENABLE);
        }
        // 礼品数量不够
        if (rewardDO.getQuantityRemain() < redeemRewardReqVO.getRewardCount()) {
            throw exception(REWARD_COUNT_NOT_ENOUGH);
        }
        // 兑换方式不匹配
        String pickMethod = rewardDO.getPickMethod();
        String[] split = pickMethod.split(",");
        if (!Arrays.asList(split).contains(String.valueOf(redeemRewardReqVO.getRedeemType()))) {
            throw exception(REWARD_PICK_METHOD_NOT_ALLOW_CREATE);
        }
        verifyMemberUser(redeemRewardReqVO, rewardDO);
        boolean lock = false;
        try {
            lock = redisDistributedLock.lock("reward:redeem:lock:" + redeemRewardReqVO.getRewardId());
            if (!lock) {
                throw exception(GET_LOCK_FAILED);
            }
            // 更新礼品
            redeemReward(rewardDO, redeemRewardReqVO.getRewardCount());
            // 添加兑换记录
            Long redeemId = addRedeemRecord(redeemRewardReqVO);
            // 更新会员积分
            updateMemberScore(redeemRewardReqVO, rewardDO, redeemId);
        } catch (Exception e) {
            throw new RuntimeException(e);
        } finally {
            redisDistributedLock.releaseLock("reward:redeem:lock:" + redeemRewardReqVO.getRewardId());
        }
        return null;
    }

    private void redeemReward(RewardDO reward, Integer rewardCount) {
        // 更新时校验礼品数量是否足够
        RewardDO rewardDO = rewardMapper.selectById(reward.getId());
        if (rewardDO.getQuantityRemain() < rewardCount) {
            throw exception(REWARD_COUNT_NOT_ENOUGH);
        }
        reward.setExchangeCount(reward.getExchangeCount() + 1);
        reward.setQuantityRemain(reward.getQuantityRemain() - rewardCount);
        rewardMapper.updateById(reward);
    }

    private void updateMemberScore(RedeemRewardReqVO redeemRewardReqVO, RewardDO rewardDO, Long redeemId) {
        Map<String, Object> extParam = new HashMap<>();
        extParam.put("redeemId", redeemId);
        memberUserScoreApi.operateScore(MemberUserScoreOperateReqDTO.builder()
                .memberId(redeemRewardReqVO.getMemberId())
                .sourceType(ScoreSourceTypeEnum.EXCHANGE_REWARD)
                .scoreCount(redeemRewardReqVO.getRewardCount() * rewardDO.getPointsRequire())
                .extParam(extParam)
                .build());
    }

    private Long addRedeemRecord(RedeemRewardReqVO redeemRewardReqVO) {
        RewardRedeemDO rewardRedeemDO = BeanUtil.copyProperties(redeemRewardReqVO, RewardRedeemDO.class);
        rewardRedeemDO.setId(snowflakeGenerator.next());
        rewardRedeemDO.setStatus(RewardRedeemStatusEnum.REDEEMING.getValue());
        rewardRedeemMapper.insert(rewardRedeemDO);
        return rewardRedeemDO.getId();
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public List<RedeemRewardRespDTO> redeemRewards(List<RedeemRewardReqVO> redeemRewardReqVOList) {
        // 批量兑换每次最多十条
        if (redeemRewardReqVOList.size() > 10) {
            throw exception(REWARD_REDEEM_COUNT_NOT_ALLOW);
        }
        Long rewardId = redeemRewardReqVOList.get(0).getRewardId();
        // 查询礼品
        RewardDO rewardDO = rewardMapper.selectById(rewardId);
        // 礼品未开启
        if (rewardDO.getStatus() != 1) {
            throw exception(REWARD_NOT_ENABLE);
        }
        int totalCount = 0;
        for (RedeemRewardReqVO redeemRewardReqVO : redeemRewardReqVOList) {
            // 每个兑换VO校验一遍
            // 兑换方式不匹配
            String pickMethod = rewardDO.getPickMethod();
            String[] split = pickMethod.split(",");
            if (!Arrays.asList(split).contains(String.valueOf(redeemRewardReqVO.getRedeemType()))) {
                throw exception(REWARD_PICK_METHOD_NOT_ALLOW_CREATE);
            }
            verifyMemberUser(redeemRewardReqVO, rewardDO);
            // 记录兑换总数
            totalCount += redeemRewardReqVO.getRewardCount();
            // 判断兑换总数是否大于礼物数量
            if (totalCount > rewardDO.getQuantityRemain()) {
                throw exception(REWARD_COUNT_NOT_ENOUGH);
            }
        }
        boolean lock = false;
        try {
            lock = redisDistributedLock.lock("reward:redeem:lock:" + rewardId);
            if (!lock) {
                throw exception(GET_LOCK_FAILED);
            }
            for (RedeemRewardReqVO redeemRewardReqVO : redeemRewardReqVOList) {
                // 更新礼品
                redeemReward(rewardDO, redeemRewardReqVO.getRewardCount());
                // 添加兑换记录
                Long redeemId = addRedeemRecord(redeemRewardReqVO);
                // 更新会员积分
                updateMemberScore(redeemRewardReqVO, rewardDO, redeemId);
            }
        } catch (Exception e) {
            throw exception(REWARD_REDEEM_FAIL);
        } finally {
            redisDistributedLock.releaseLock("reward:redeem:lock:" + rewardId);
        }
        return null;
    }

    private void verifyMemberUser(RedeemRewardReqVO redeemRewardReqVO, RewardDO rewardDO) {
        // 查询会员积分
        UserRespDTO memberUser = memberUserApi.getUser(redeemRewardReqVO.getMemberId());
        if (memberUser == null) {
            throw exception(USER_NOT_EXISTS);
        }
        Integer holdScore = memberUser.getHoldScore();
        // 会员积分不够
        if (holdScore < rewardDO.getPointsRequire() * redeemRewardReqVO.getRewardCount()) {
            throw exception(REWARD_SCORE_NOT_ENOUGH);
        }
        // 校验兑换次数
        LambdaQueryWrapper<RewardRedeemDO> rewardRedeemDOWrapper = new LambdaQueryWrapper<>();
        rewardRedeemDOWrapper.eq(RewardRedeemDO::getRewardId, redeemRewardReqVO.getRewardId())
                .eq(RewardRedeemDO::getMemberId, redeemRewardReqVO.getMemberId());
        Long count = rewardRedeemMapper.selectCount(rewardRedeemDOWrapper);
        if (count > rewardDO.getAllowCount()) {
            throw exception(REWARD_REDEEM_ALLOW_COUNT_ERROR);
        }
    }
}
